import { LogFlagList, TokenList, SpanParseNode } from './DataTypes'
import { InsideConsoleLog }                      from './Log'
import { SpanParserQueue }                       from './SpanParserQueue'
import   Lexer                                   from './Lexer'

/**
 *  **Span Parser Class**
 * 
 * The most important class of the application. This parser is very simples, and
 * `only can be process HTML tag span WITH class`. It is using a QUEUE and not a 
 * tree. It means that it does not process tag within tag. Just parse the format
 * returned  by  HIGHLIGHT.JS  lib.  If your code returned fail,  means it has a 
 * syntax error in your code. Until it becomes a Parse Tree Compiler, there will
 * be this problem.  
 * ---
 * Does this mean the module is useless? `NO`. It helps you to prevent the XSS 
 * vunerability attacks. Avoiding using the function `dangerouslySetInnerHTML`.
 *  
 * https://owasp.org/www-community/attacks/xss/
 * 
 * ---
 * ```typescript
 * 
 * CLASS
 *   Scan      : Lexer            // Lexer/HljsToReact Class 
 *   SpanASQ   : SpanParserQueue  // Abstract Syntax Queue
 *   Log       : InsideConsoleLog // Only verbose Log in terminal
 *   tagContent: string           // Buffer for tag content
 *   txtContent: string           // Buffer for class text content
 * ```
 */

export default 
class Parser {
    Scan      : Lexer 
    SpanASQ   : SpanParserQueue
    Log       : InsideConsoleLog
    tagContent = ''
    txtContent = ''

    /**
     * 
     * ```typescript
     * constructor 
     *   >> Lexer Class
     * ```
     */
    constructor({ Lexer }: { Lexer: Lexer }){
        this.Scan    = Lexer
        this.SpanASQ = new SpanParserQueue()
        this.Log     = new InsideConsoleLog()
    }

    /**
     * The main function of the Parser. It contains the logic to detect Span tag
     * with className(class).
     * 
     * **SYNTAX**
     * 
     * ```typescript
     * CLASS_TEXT
     *    'class' '=' '\"' all text in '\"'
     * CONTENT
     *    '>' all text in '<'
     * TAG
     *    CONTENT '<' 'span' CLASS_TEXT '>' CONTENT '<' '/' 'span' '>' CONTENT
     * ```
     */
    exec(): LogFlagList._ok | LogFlagList._parser_error {
        let status  = 0
        
        this.Log.initSimpleCompiler()

        do {      
            this.tagContent = this.Scan.content
            
            status = this.openingSpan()
            if( status === LogFlagList._parser_error || 
                status === LogFlagList._Lexer_error ) {
                console.log( 'Parser Failed in ::' )
                console.log( '--- openingSpan( ) ')
                return LogFlagList._parser_error
            }

            this.tagContent = this.filter( this.Scan.content )

            status = this.closingSpan()
            if( status === LogFlagList._parser_error || 
                status === LogFlagList._Lexer_error ) {
                console.log( 'Parser Failed in ::' )
                console.log( '--- closingSpan( ) ')
                return LogFlagList._parser_error
            }

            this.SpanASQ.add( new SpanParseNode(
                TokenList._span,
                this.tagContent,
                this.txtContent,
                )
            )

            this.clearBuffer()

            this.tagContent = this.filter( this.Scan.content )

            if( this.tagContent !== '') {
                this.SpanASQ.add( new SpanParseNode(
                    TokenList._content,
                    this.tagContent,
                    )
                )   
            }

            this.clearBuffer()

        } while( this.Scan.checkEof() )
        
        this.Log.endSimpleCompiler()

        return LogFlagList._ok
    }

    /**
     * Receive a string HTML entities and convert to plain/text
     * ```typescript
     * filter
     *   >> txt         : string 
     *   << filterd_txt : string
     * ```
     */
    filter( txt: string ) : string {
        return   txt.replace(/&lt;/g   , '<' )
                    .replace(/&gt;/g   , '>' )
                    .replace(/&amp;/g  , '&' )
                    .replace(/&#39;/g  , '\'') 
                    .replace(/&#x27;/g , '\'') 
                    .replace(/&quot;/g , '"' )  
    }

    perror(msg: string, error_type: number) {
        if( error_type === LogFlagList._Lexer_error )
            console.log('Token Error::')
        else
        if( error_type === LogFlagList._parser_error )
            console.log('Parser Error')
        console.log('- - - ' +msg)
    }

    /**
     * Only clean the buffer content for each new child
     */
    clearBuffer(): void {
        this.tagContent = ''
        this.txtContent = ''
    }
 
    /**
     * Search for the Opening Span Tag 
     * `< span class=" txtContent " >`
     * 
     * ```typescript
     * openingSpan
     *   << LogFlagList.(_ok | _error)
     * ```
     */
    openingSpan(): LogFlagList._ok | LogFlagList._Lexer_error {
        this.Log.initParser('Opening Span Tag')

        // IF token '<' 
        if(this.Scan.token() !== '<'.charCodeAt(0)){
            this.perror( 
                'Undefined Token (' + this.Scan.lastChar() + ')',
                LogFlagList._Lexer_error 
            )

            return LogFlagList._Lexer_error
        }        
        this.Log.endLexer()
        
        // IF token 'span'
        if( this.Scan.token() !== TokenList._span ) {
            this.perror( 
                'Undefined Token (' + this.Scan.lastChar() + ')',
                LogFlagList._Lexer_error 
            )
            return LogFlagList._Lexer_error
        }
        this.Log.endLexer()

        // IF CLASS_TEXT
        let  last_token = this.Scan.token() 
        this.Log.endLexer()

          // The format is prepared to use empty class span, but not yet used
        if(  last_token    === TokenList._class ) {
            if( this.Scan.token() !== '='.charCodeAt(0) ) {
                this.perror( 
                    'Undefined Token (' + this.Scan.lastChar() + ')',
                    LogFlagList._Lexer_error 
                )
                return LogFlagList._Lexer_error
            }
            this.Log.endLexer()

            if( this.Scan.token() !== '"'.charCodeAt(0) ) {
                this.perror( 
                    'Undefined Token (' + this.Scan.lastChar() + ')',
                    LogFlagList._Lexer_error 
                )
                return LogFlagList._Lexer_error
            }
            this.Log.endLexer()

            // Get the text contente inside '\"' 
            this.txtContent = this.Scan.content

            // If token '\"'
            if( this.Scan.token() !== '"'.charCodeAt(0) ) {
                this.perror( 
                    'Undefined Token (' + this.Scan.lastChar() + ')',
                    LogFlagList._Lexer_error 
                )
                return LogFlagList._Lexer_error
            }
            this.Log.endLexer()
        
            // Token Buffer to use with not empty class span only
            last_token = this.Scan.token()
        }
        
        // IF token '>' : closing correctly
        if( last_token !== '>'.charCodeAt(0) ) {
            this.perror( 
                'Undefined Token (' + this.Scan.lastChar() + ')',
                LogFlagList._Lexer_error 
            )
            return LogFlagList._Lexer_error
        }   
        this.Log.endLexer()
        this.Log.endParser({ name: 'Opening Span Tag' })

        return LogFlagList._ok
    }

    /**
     * Search for the Closing Span Tag 
     * `< / span >`
     * 
     * ```typescript
     * closingSpan
     *   << LogFlagList.(_ok | _error)
     * ```
     */
    closingSpan(): LogFlagList._ok | LogFlagList._Lexer_error {
        this.Log.initParser('Closing Span Tag')

        //IF token '<'
        if(this.Scan.token() !== '<'.charCodeAt(0)){
            this.perror( 
                'Undefined Token (' + this.Scan.lastChar() + ')',
                LogFlagList._Lexer_error 
            )
            return LogFlagList._Lexer_error
        }
        this.Log.endLexer()

        //IF token '/'
        if(this.Scan.token() !== '/'.charCodeAt(0)){
            this.perror( 
                'Undefined Token (' + this.Scan.lastChar() + ')',
                LogFlagList._Lexer_error 
            )
            return LogFlagList._Lexer_error
        }
        this.Log.endLexer()

        //IF token 'span'
        if( this.Scan.token() !== TokenList._span ) {
            this.perror( 
                'Undefined Token (' + this.Scan.lastChar() + ')',
                LogFlagList._Lexer_error 
            )
            return LogFlagList._Lexer_error
        }
        this.Log.endLexer()

        //IF token '>' : closing correctly
        if( this.Scan.token() !== '>'.charCodeAt(0) ) {
            this.perror( 
                'Undefined Token (' + this.Scan.lastChar() + ')',
                LogFlagList._Lexer_error 
            )
            return LogFlagList._Lexer_error
        } 
        this.Log.endLexer()
        this.Log.endParser({ name: 'Closing Span Tag' })

        return LogFlagList._ok
    }
}